@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/child.c,v 1.26.2.1 2009/05/02 08:41:03 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/child.c,v 1.26.2.4 2009/08/01 11:40:38 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -60,7 +60,6 @@
 static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *timeout);
 static StartupPacket *read_startup_packet(POOL_CONNECTION *cp);
 static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION *frontend);
-static void cancel_request(CancelPacket *sp);
 static RETSIGTYPE die(int sig);
 static RETSIGTYPE close_idle_connection(int sig);
 static RETSIGTYPE wakeup_handler(int sig);
@@ -1079,7 +1078,7 @@ int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp)
 /*
  * process cancel request
  */
-static void cancel_request(CancelPacket *sp)
+void cancel_request(CancelPacket *sp)
 {
 	int	len;
 	int fd;
@@ -1250,6 +1249,8 @@ static RETSIGTYPE close_idle_connection(int sig)
 	{
 		if (!MASTER_CONNECTION(p))
 			continue;
+ 		if (!MASTER_CONNECTION(p)->sp)
+			continue;
 		if (MASTER_CONNECTION(p)->sp->user == NULL)
 			continue;
 
@@ -1310,6 +1311,8 @@ static void send_frontend_exits(void)
 	{
 		if (!MASTER_CONNECTION(p))
 			continue;
+		if (!MASTER_CONNECTION(p)->sp)
+			continue;
 		if (MASTER_CONNECTION(p)->sp->user == NULL)
 			continue;
 		pool_send_frontend_exits(p);
@@ -1654,7 +1657,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	}
 	else
 	{
-		pool_error("s_do_auth: auth kind %d not supported yet");
+		pool_error("s_do_auth: auth kind %d not supported yet", auth_kind);
 		return -1;
 	}
 
