@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pool_connection_pool.c,v 1.9 2007/07/26 10:15:45 y-asaba Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/pool_connection_pool.c,v 1.13.2.2 2009/08/11 13:23:52 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2007	PgPool Global Development Group
+ * Copyright (c) 2003-2009	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -101,6 +101,7 @@ POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor, in
 	for (i=0;i<pool_config->max_pool;i++)
 	{
 		if (MASTER_CONNECTION(p) &&
+			MASTER_CONNECTION(p)->sp &&
 			MASTER_CONNECTION(p)->sp->major == protoMajor &&
 			MASTER_CONNECTION(p)->sp->user != NULL &&
 			strcmp(MASTER_CONNECTION(p)->sp->user, user) == 0 &&
@@ -120,9 +121,17 @@ POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor, in
 					if (!VALID_BACKEND(j))
 						continue;
 
-					sock_broken = check_socket_status(CONNECTION(p, j)->fd);
-					if (sock_broken < 0)
+					if  (CONNECTION_SLOT(p, j))
+					{
+						sock_broken = check_socket_status(CONNECTION(p, j)->fd);
+						if (sock_broken < 0)
+							break;
+					}
+					else
+					{
+						sock_broken = -1;
 						break;
+					}
 				}
 
 				if (sock_broken < 0)
@@ -130,7 +139,7 @@ POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor, in
 					pool_log("connection closed. retry to create new connection pool.");
 					for (j=0;j<NUM_BACKENDS;j++)
 					{
-						if (!VALID_BACKEND(j))
+						if (!VALID_BACKEND(j) || (CONNECTION_SLOT(p, j) == NULL))
 							continue;
 
 						if (!freed)
@@ -146,9 +155,11 @@ POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor, in
 					memset(p, 0, sizeof(POOL_CONNECTION_POOL_SLOT));
 					p->info = info;
 					memset(p->info, 0, sizeof(ConnectionInfo));
+					POOL_SETMASK(&oldmask);
 					return NULL;
 				}
 			}
+			POOL_SETMASK(&oldmask);
 			return p;
 		}
 		p++;
@@ -228,7 +239,7 @@ POOL_CONNECTION_POOL *pool_create_cp(void)
 	closetime = MASTER_CONNECTION(p)->closetime;
 	for (i=0;i<pool_config->max_pool;i++)
 	{
-		pool_debug("user: %s database: %s closetime: %d",
+		pool_debug("user: %s database: %s closetime: %ld",
 				   MASTER_CONNECTION(p)->sp->user,
 				   MASTER_CONNECTION(p)->sp->database,
 				   MASTER_CONNECTION(p)->closetime);
@@ -279,7 +290,7 @@ void pool_connection_pool_timer(POOL_CONNECTION_POOL *backend)
 	POOL_CONNECTION_POOL *p = pool_connection_pool;
 	int i;
 
-	pool_debug("pool_connection_pool_timer: set close time %d", time(NULL));
+	pool_debug("pool_connection_pool_timer: set close time %ld", time(NULL));
 
 	MASTER_CONNECTION(backend)->closetime = time(NULL);		/* set connection close time */
 
@@ -291,6 +302,8 @@ void pool_connection_pool_timer(POOL_CONNECTION_POOL *backend)
 	{
 		if (!MASTER_CONNECTION(p))
 			continue;
+		if (!MASTER_CONNECTION(p)->sp)
+			continue;
 		if (MASTER_CONNECTION(p)->sp->user == NULL)
 			continue;
 
@@ -326,12 +339,14 @@ void pool_backend_timer(void)
 
 	now = time(NULL);
 
-	pool_debug("pool_backend_timer_handler called at %d", now);
+	pool_debug("pool_backend_timer_handler called at %ld", now);
 
 	for (i=0;i<pool_config->max_pool;i++, p++)
 	{
 		if (!MASTER_CONNECTION(p))
 			continue;
+		if (!MASTER_CONNECTION(p)->sp)
+			continue;
 		if (MASTER_CONNECTION(p)->sp->user == NULL)
 			continue;
 
@@ -340,7 +355,7 @@ void pool_backend_timer(void)
 		{
 			int freed = 0;
 
-			pool_debug("pool_backend_timer_handler: expire time: %d",
+			pool_debug("pool_backend_timer_handler: expire time: %ld",
 					   MASTER_CONNECTION(p)->closetime+pool_config->connection_life_time);
 
 			if (now >= (MASTER_CONNECTION(p)->closetime+pool_config->connection_life_time))
@@ -555,8 +570,7 @@ static POOL_CONNECTION_POOL *new_connection(POOL_CONNECTION_POOL *p)
 	{
 		pool_debug("new_connection: connecting %d backend", i);
 
-		if (BACKEND_INFO(i).backend_status == CON_DOWN ||
-			BACKEND_INFO(i).backend_status == CON_UNUSED)
+		if (!VALID_BACKEND(i))
 		{
 			pool_debug("new_connection: skipping slot %d because backend_status = %d",
 					   i, BACKEND_INFO(i).backend_status);
@@ -570,8 +584,6 @@ static POOL_CONNECTION_POOL *new_connection(POOL_CONNECTION_POOL *p)
 			return NULL;
 		}
 
-		p->slots[i] = s;
-
 		if (create_cp(s, i) == NULL)
 		{
 			/* connection failed. mark this backend down */
@@ -583,6 +595,8 @@ static POOL_CONNECTION_POOL *new_connection(POOL_CONNECTION_POOL *p)
 			child_exit(1);
 		}
 
+		p->slots[i] = s;
+
 		if (pool_init_params(&s->con->params))
 		{
 			return NULL;
