@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
-* $Header: /cvsroot/pgpool/pgpool-II/pool_stream.c,v 1.6 2007/11/14 03:59:58 y-mori Exp $
+* $Header: /cvsroot/pgpool/pgpool-II/pool_stream.c,v 1.14.2.1 2009/05/02 08:45:35 t-ishii Exp $
 *
 * pgpool: a language independent connection pool server for PostgreSQL 
 * written by Tatsuo Ishii
 *
-* Copyright (c) 2003-2007	PgPool Global Development Group
+* Copyright (c) 2003-2009	PgPool Global Development Group
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
@@ -95,7 +95,13 @@ POOL_CONNECTION *pool_open(int fd)
 */
 void pool_close(POOL_CONNECTION *cp)
 {
+	/*
+	 * shutdown connection to the client so that pgpool is not blocked
+	 */
+	if (!cp->isbackend)
+		shutdown(cp->fd, 1);
 	close(cp->fd);
+
 	free(cp->wbuf);
 	free(cp->hp);
 	if (cp->sbuf)
@@ -116,18 +122,14 @@ int pool_read(POOL_CONNECTION *cp, void *buf, int len)
 
 	int consume_size;
 	int readlen;
-	int notimeout;
 
 	consume_size = consume_pending_data(cp, buf, len);
 	len -= consume_size;
 	buf += consume_size;
 
-	/* if this is not the master backend, then set timeout */
-	notimeout = !IS_MASTER_NODE_ID(cp->db_node_id);
-
 	while (len > 0)
 	{
-		if (pool_check_fd(cp, notimeout))
+		if (pool_check_fd(cp))
 		{
 			if (!IS_MASTER_NODE_ID(cp->db_node_id))
 			{
@@ -168,105 +170,7 @@ int pool_read(POOL_CONNECTION *cp, void *buf, int len)
 		{
 			if (cp->isbackend)
 			{
-				pool_error("pool_read2: EOF encountered with backend");
-				return -1;
-
-#ifdef NOT_USED
-			    /* fatal error, notice to parent and exit */
-			    notice_backend_error(IS_MASTER_NODE_ID(cp->db_node_id));
-				child_exit(1);
-#endif
-			}
-			else
-			{
-				/*
-				 * if backend offers authentication method, frontend could close connection
-				 */
-				return -1;
-			}
-		}
-
-		if (len < readlen)
-		{
-			/* overrun. we need to save remaining data to pending buffer */
-			if (save_pending_data(cp, readbuf+len, readlen-len))
-				return -1;
-			memmove(buf, readbuf, len);
-			break;
-		}
-
-		memmove(buf, readbuf, readlen);
-		buf += readlen;
-		len -= readlen;
-	}
-
-	return 0;
-}
-
-/*
-* this function is used by SELECT QUERY at parallel query
-* read len bytes from cp
-* returns 0 on success otherwise -1.
-*/
-int pool_read_parallel(POOL_CONNECTION *cp, void *buf, int len)
-{
-	static char readbuf[READBUFSZ];
-
-	int consume_size;
-	int readlen;
-	int notimeout;
-
-	consume_size = consume_pending_data(cp, buf, len);
-	len -= consume_size;
-	buf += consume_size;
-
-	/* if this is not the master backend, then set timeout */
-	notimeout = 1;
-
-	while (len > 0)
-	{
-		if (pool_check_fd(cp, notimeout))
-		{
-			if (!IS_MASTER_NODE_ID(cp->db_node_id))
-			{
-				pool_log("pool_read: data is not ready in DB node: %d. abort this session",
-						 cp->db_node_id);
-				exit(1);
-			}
-			else
-			{
-				pool_error("pool_read: pool_check_fd failed (%s)", strerror(errno));
-			    return -1;
-			}
-		}
-
-		readlen = read(cp->fd, readbuf, READBUFSZ);
-		if (readlen == -1)
-		{
-			if (errno == EINTR || errno == EAGAIN)
-			{
-				pool_debug("pool_read: retrying due to %s", strerror(errno));
-				continue;
-			}
-
-			pool_error("pool_read: read failed (%s)", strerror(errno));
-
-			if (cp->isbackend)
-			{
-			    /* fatal error, notice to parent and exit */
-				notice_backend_error(cp->db_node_id);
-				child_exit(1);
-			}
-			else
-			{
-			    return -1;
-			}
-		}
-		else if (readlen == 0)
-		{
-			if (cp->isbackend)
-			{
-				pool_error("pool_read2: EOF encountered with backend");
+				pool_error("pool_read: EOF encountered with backend");
 				return -1;
 
 #ifdef NOT_USED
@@ -312,7 +216,6 @@ char *pool_read2(POOL_CONNECTION *cp, int len)
 	int alloc_size;
 	int consume_size;
 	int readlen;
-	int notimeout;
 
 	req_size = cp->len + len;
 
@@ -334,12 +237,9 @@ char *pool_read2(POOL_CONNECTION *cp, int len)
 	len -= consume_size;
 	buf += consume_size;
 
-	/* if this is not the master backend, then set timeout */
-	notimeout = IS_MASTER_NODE_ID(cp->db_node_id);
-
 	while (len > 0)
 	{
-		if (pool_check_fd(cp, notimeout))
+		if (pool_check_fd(cp))
 		{
 			if (!IS_MASTER_NODE_ID(cp->db_node_id))
 			{
@@ -411,8 +311,6 @@ char *pool_read2(POOL_CONNECTION *cp, int len)
 */
 int pool_write(POOL_CONNECTION *cp, void *buf, int len)
 {
-	int reqlen;
-
 	if (len < 0)
 	{
 		pool_error("pool_write: invalid request size: %d", len);
@@ -422,26 +320,32 @@ int pool_write(POOL_CONNECTION *cp, void *buf, int len)
 	if (cp->no_forward)
 		return 0;
 
-	/* check buffer size */
-	reqlen = cp->wbufpo + len;
-
-	if (reqlen > cp->wbufsz)
+	while (len > 0)
 	{
-		char *p;
+		int remainder = WRITEBUFSZ - cp->wbufpo;
 
-		reqlen = (reqlen/WRITEBUFSZ+1)*WRITEBUFSZ;
-		p = realloc(cp->wbuf, reqlen);
-		if (p == NULL)
+		if (cp->wbufpo >= WRITEBUFSZ)
 		{
-			pool_error("pool_write: realloc failed");
-			return -1;
+			/*
+			 * Write buffer is full. so flush buffer.
+			 * wbufpo is reset in pool_flush_it().
+			 */
+			if (pool_flush_it(cp) == -1)
+				return -1;
+			remainder = WRITEBUFSZ;
 		}
-		cp->wbuf = p;
-		cp->wbufsz = reqlen;
-	}
 
-	memcpy(cp->wbuf+cp->wbufpo, buf, len);
-	cp->wbufpo += len;
+		/* check buffer size */
+		if (remainder >= len)
+		{
+			/* OK, buffer size is enough. */
+			remainder = len;
+		}
+		memcpy(cp->wbuf+cp->wbufpo, buf, remainder);
+		cp->wbufpo += remainder;
+		buf += remainder;
+		len -= remainder;
+	}
 
 	return 0;
 }
@@ -485,6 +389,7 @@ int pool_flush_it(POOL_CONNECTION *cp)
 					continue;
 
 				pool_error("pool_flush_it: select() failed. reason: %s", strerror(errno));
+				cp->wbufpo = 0;
 				return -1;
 			}
 			else if (sts == 0)
@@ -494,6 +399,7 @@ int pool_flush_it(POOL_CONNECTION *cp)
 			else if (FD_ISSET(cp->fd, &exceptmask))
 			{
 				pool_log("pool_flush_it: exception occured");
+				cp->wbufpo = 0;
 				return -1;
 			}
 		}
@@ -513,6 +419,7 @@ int pool_flush_it(POOL_CONNECTION *cp)
 			else if (wlen < 0)
 			{
 				pool_error("pool_flush_it: invalid write size %d", sts);
+				cp->wbufpo = 0;
 				return -1;
 			}
 
@@ -533,6 +440,7 @@ int pool_flush_it(POOL_CONNECTION *cp)
 		{
 			pool_error("pool_flush_it: write failed (%s) offset: %d wlen: %d",
 					   strerror(errno), offset, wlen);
+			cp->wbufpo = 0;
 			return -1;
 		}
 	}
@@ -556,7 +464,11 @@ int pool_flush(POOL_CONNECTION *cp)
 		}
 		else
 		{
-			return -1;
+			/*
+			 * ignore error on frontend. we need to continue the
+			 * processing with backends
+			 */
+			return 0;
 		}
 	}
 	return 0;
@@ -584,7 +496,6 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 	int strlength;
 	int flag;
 	int consume_size;
-	int notimeout;
 
 #ifdef DEBUG
 	static char pbuf[READBUFSZ];
@@ -651,12 +562,9 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 		readsize = cp->sbufsz;
 	}
 
-	/* if this is not the master backend, then set timeout */
-	notimeout = IS_MASTER_NODE_ID(cp->db_node_id);
-
 	for (;;)
 	{
-		if (pool_check_fd(cp, notimeout))
+		if (pool_check_fd(cp))
 		{
 			if (!IS_MASTER_NODE_ID(cp->db_node_id))
 			{
